/// dart_result Is a library that exports the [Result] type.
///
/// The primary use case for this library is to enable representation of potential failures
/// in code without resorting to [Exception]s.
/// To represent a failure in a code path, use the [Failure] function.
/// To represent success, use the [Success] function.
/// [Result] has various methods that allow you to check if the Result was a Success or a Failure.
library dart_result;

import 'package:union/union.dart';

/// The result of an operation that may result in an error.
class Result<T> extends Union2<T, String> {
  /// Constructor for the success case.
  const Result.success(T value) : super.first(value);

  /// Constructor for the failure case.
  const Result.failure(String cause) : super.second(cause);

  /// Returns the result of applying [mapping] to [T] if the Result is success.
  ///
  /// If the result is failure, returns the failure cause unchanged.
  Result<E> mapSuccess<E>(E Function(T) mapping) =>
      isFailure ? Failure(this.value) : Success(mapping(this.value));

  /// Returns true if the result is success.
  bool get isSuccess => this.value is T;

  /// Returns true if the result is failure.
  bool get isFailure => !isSuccess;

  /// Returns the result of applying either [onSuccess] or [onFailure] based on
  ///
  /// whether the result [isSuccess] or [isFailure].
  E match<E>(E Function(T) onSuccess, E Function(String) onFailure) =>
      isSuccess ? onSuccess(this.value) : onFailure(this.value);

  /// Returns the inner value [T] if this [isSuccess], otherwise applies the [onFailure]
  ///
  /// function to the failure string and returns a value of [T].
  T getOrElse(T Function(String) onFailure) =>
      isSuccess ? this.value : onFailure(this.value);

  @override
  String toString() {
    return isSuccess
        ? 'Success{value: ${this.value}}'
        : 'Failure{cause: ${this.value} }';
  }

  @override
  bool operator ==(Object other) => other is Result && value == other.value;

  @override
  int get hashCode => this?.value?.hashCode;
}

/// Creates a successful [Result].
Result<T> Success<T>(T value) => Result.success(value);

/// Creates a failed [Result].
Result<T> Failure<T>(String cause) => Result.failure(cause);
